/*******************************************************************************
 * $Id: Islands.cpp,v 1.12 2004/04/19 23:11:19 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "libMems/Islands.h"
#include "libMems/Aligner.h"

/**
 * Identifies gaps in the alignment between pairs of sequences that are longer than
 * some number of base pairs in length.  Prints islands to an output stream
 */
void simpleFindIslands( IntervalList& iv_list, uint island_size, ostream& island_out ){
	if( iv_list.size() == 0 )
		return;
	for( uint iv_listI = 0; iv_listI < iv_list.size(); iv_listI++ ){
		Interval& iv = iv_list[ iv_listI ];
		gnAlignedSequences gnas;
		iv.GetAlignedSequences( gnas, iv_list.seq_table );
		uint seq_count = iv_list.seq_table.size();
		
		for( uint seqI = 0; seqI < seq_count; seqI++ ){
			uint seqJ;
			for( seqJ = seqI + 1; seqJ < seq_count; seqJ++ ){
				uint columnI = 0;
				gnSeqI curI = 0;
				gnSeqI curJ = 0;
				gnSeqI lastI = 0;
				gnSeqI lastJ = 0;
				for( columnI = 0; columnI < gnas.alignedSeqsSize(); columnI++ ){
					if( gnas.sequences[ seqI ][ columnI ] != '-' )
						curI++;
					if( gnas.sequences[ seqJ ][ columnI ] != '-' )
						curJ++;
					if( toupper( gnas.sequences[ seqI ][ columnI ] ) == 
						toupper( gnas.sequences[ seqJ ][ columnI ] ) &&
						gnas.sequences[ seqJ ][ columnI ] != '-' ){
						// check for an island that was big enough
						if( curI - lastI > island_size ||
							curJ - lastJ > island_size ){
							int64 leftI = iv.Start( seqI );
							int64 rightI = leftI < 0 ? leftI - curI : leftI + curI;
							leftI = leftI < 0 ? leftI - lastI : leftI + lastI;
							int64 leftJ = iv.Start( seqJ );
							int64 rightJ = leftJ < 0 ? leftJ - curJ : leftJ + curJ;
							leftJ = leftJ < 0 ? leftJ - lastJ : leftJ + lastJ;
							island_out << seqI << '\t' << leftI << '\t' << rightI << '\t' 
									<< seqJ << '\t' << leftJ << '\t' << rightJ << endl;
						}
						
						lastI = curI;
						lastJ = curJ;
					}
				}
			}
		}
	}
}


/**
 * Identifies stretches of alignment existing in all sequences that doesn't
 * contain a gap larger than a particular size.  Such regions are considered
 * the backbone of the alignment.
 */
void simpleFindBackbone( IntervalList& iv_list, uint backbone_size, uint max_gap_size, vector< GappedAlignment >& backbone_regions ){
	if( iv_list.size() == 0 )
		return;
	for( uint iv_listI = 0; iv_listI < iv_list.size(); iv_listI++ ){
		Interval& iv = iv_list[ iv_listI ];
		gnAlignedSequences gnas;
		uint seqI;
		uint seq_count = iv_list.seq_table.size();
		vector< int64 > positions( seq_count );
		vector< int64 > starts( seq_count );
		vector< int64 > ends( seq_count );
		vector< uint > gap_size( seq_count, 0 );
		uint seqJ;
		gnSeqI bb_start_col = 0;
		gnSeqI bb_end_col = 0;
		GappedAlignment cur_backbone( seq_count, 0 );
		
		// initialize positions and starts
		for( seqI = 0; seqI < seq_count; seqI++ ){
			positions[ seqI ] = iv_list[ iv_listI ].Start( seqI );
			if( positions[ seqI ] < 0 )
				positions[ seqI ] -= iv_list[ iv_listI ].Length( seqI ) + 1;
		}
		starts = positions;
		ends = positions;

		iv.GetAlignedSequences( gnas, iv_list.seq_table );
		bool backbone = true;	// assume we are starting out with a complete alignment column
		uint columnI = 0;
		vector< int64 > prev_positions;
		for( ; columnI < gnas.alignedSeqsSize(); columnI++ ){
			bool no_gaps = true;
			prev_positions = positions;
			for( seqI = 0; seqI < seq_count; seqI++ ){
				if( gnas.sequences[ seqI ][ columnI ] != '-' ){
					if( gap_size[ seqI ] > max_gap_size && backbone ){
						// end a stretch of backbone here only
						// if the backbone meets size requirements in each
						// sequence.
						for( seqJ = 0; seqJ < seq_count; seqJ++ ){
							if( ends[ seqJ ] - starts[ seqJ ] < backbone_size ){
								break;
							}
						}
						if( seqJ == seq_count ) {
							// it's a legitimate stretch of backbone
							backbone_regions.push_back( cur_backbone );
							uint bbI = backbone_regions.size() - 1;
							for( seqJ = 0; seqJ < seq_count; seqJ++ ){
								if( starts[ seqJ ] < 0 )
									backbone_regions[ bbI ].SetStart( seqJ, ends[ seqJ ] + 1);
								else
									backbone_regions[ bbI ].SetStart( seqJ, starts[ seqJ ] );
								backbone_regions[ bbI ].SetLength( ends[ seqJ ] - starts[ seqJ ], seqJ );
								backbone_regions[ bbI ].align_matrix[ seqJ ] = gnas.sequences[ seqJ ].substr( bb_start_col, bb_end_col - bb_start_col + 1);
								backbone_regions[ bbI ].align_length = bb_end_col - bb_start_col + 1;
							}
							
						}
						// we either just finished backbone or a short area that didn't
						// qualify as backbone
						// look for a new backbone region
						backbone = false;
					}
					positions[ seqI ]++;
					gap_size[ seqI ] = 0;
				}else{
					gap_size[ seqI ]++;
					no_gaps = false;
				}
			}
			if( no_gaps ){
				bb_end_col = columnI;
				ends = positions;
				if( !backbone ){
					starts = prev_positions;
					bb_start_col = columnI;
					backbone = true;
				}
			}
		}

		// check for backbone one last time
		for( seqJ = 0; seqJ < seq_count; seqJ++ ){
			if( ends[ seqJ ] - starts[ seqJ ] < backbone_size ){
				break;
			}
		}
		if( seqJ == seq_count ) {
			// it's a legitimate stretch of backbone
			backbone_regions.push_back( cur_backbone );
			uint bbI = backbone_regions.size() - 1;
			for( seqJ = 0; seqJ < seq_count; seqJ++ ){
				if( starts[ seqJ ] < 0 )
					backbone_regions[ bbI ].SetStart( seqJ, ends[ seqJ ] + 1);
				else
					backbone_regions[ bbI ].SetStart( seqJ, starts[ seqJ ] );
				backbone_regions[ bbI ].SetLength( ends[ seqJ ] - starts[ seqJ ], seqJ );
				backbone_regions[ bbI ].align_matrix[ seqJ ] = gnas.sequences[ seqJ ].substr( bb_start_col, bb_end_col - bb_start_col + 1);
				backbone_regions[ bbI ].align_length = bb_end_col - bb_start_col + 1;
			}
		}
	}
}


void outputBackbone( const vector< GappedAlignment >& backbone_regions, ostream& backbone_out ){
	for( uint bbI = 0; bbI < backbone_regions.size(); bbI++ ){
		for( uint seqJ = 0; seqJ < backbone_regions[ bbI ].SeqCount(); seqJ++ ){
			if( seqJ > 0 )
				backbone_out << '\t';
			backbone_out << backbone_regions[ bbI ].Start( seqJ ) << '\t' << backbone_regions[ bbI ].Start( seqJ ) + (int64)backbone_regions[ bbI ].Length( seqJ );
		}
		backbone_out << endl;
	}
}


// always return the left end of the one to the left and the right of the one to the right

void getGapBounds( IntervalList& iv_list, vector< LCB >& adjacencies, uint seqJ, int leftI, int rightI, int64& left_start, int64& right_start );
void getGapBounds( IntervalList& iv_list, vector< LCB >& adjacencies, uint seqJ, int leftI, int rightI, int64& left_start, int64& right_start ){
	if( rightI != -1 )
		right_start = abs( adjacencies[ rightI ].left_end[ seqJ ] );
	else
		right_start = iv_list.seq_table[ seqJ ]->length() + 1;
	
	if( leftI != -1 )
		left_start = abs( adjacencies[ leftI ].right_end[ seqJ ] );
	else
		left_start = 1;
}

void addUnalignedIntervals( IntervalList& iv_list ){
	SlotAllocator< Match >& static_match_allocator = SlotAllocator< Match >::GetSlotAllocator();
	Aligner al( iv_list.seq_table.size() );
	vector< LCB > adjacencies;
	vector< int64 > weights;
	uint lcbI;
	uint seqI;
	uint seq_count = iv_list.seq_table.size();
	
	for( lcbI = 0; lcbI < iv_list.size(); lcbI++ )
		weights.push_back( 0 );

	al.computeLCBAdjacencies_v2( iv_list, weights, adjacencies );

	vector< int > rightmost;
	for( seqI = 0; seqI < seq_count; seqI++ ){
		rightmost.push_back( -1 );
	}
	for( lcbI = 0; lcbI <= adjacencies.size(); lcbI++ ){
		for( seqI = 0; seqI < seq_count; seqI++ ){
			// scan left
			int leftI;
			if( lcbI < adjacencies.size() ){
// left is always to the left!!
				leftI = adjacencies[ lcbI ].left_adjacency[ seqI ];
			}else
				leftI = rightmost[ seqI ];

			int rightI = lcbI < adjacencies.size() ? lcbI : -1;
// right is always to the right!!
			if( lcbI < adjacencies.size() )
				if( adjacencies[ lcbI ].right_adjacency[ seqI ] == -1 )
					rightmost[ seqI ] = lcbI;
			
			int64 left_start, right_start;
			getGapBounds( iv_list, adjacencies, seqI, leftI, rightI, left_start, right_start );
			int64 gap_len =  abs( right_start ) - abs( left_start );
			if( gap_len > 0 ){
				Interval iv( seq_count, gap_len );
				Match* m = static_match_allocator.Allocate();
				*m = Match( seq_count, 0, Match::extended );
				for( uint seqJ = 0; seqJ < seq_count; seqJ++ ){
					m->SetStart( seqJ, 0 );
				}
				m->SetStart( seqI, left_start );
				m->SetLength( gap_len );
				m->CalculateOffset();
				iv.add( m );
				iv.CalculateOffset();
				iv_list.push_back( iv );
			}
		}
	}
}


void findIslandsBetweenLCBs( IntervalList& iv_list, uint island_size, ostream& island_out ){
	IntervalList iv_list_tmp = iv_list;
	addUnalignedIntervals( iv_list_tmp );
	uint seq_count = iv_list.seq_table.size();
	
	for( int ivI = iv_list.size(); ivI < iv_list_tmp.size(); ivI++ ){
		for( uint seqI = 0; seqI < seq_count; seqI++ ){
			if( iv_list_tmp[ ivI ].Length( seqI ) < island_size )
				continue;

			// this is an island, write the LCB island out
			gnSeqI left_end = abs( iv_list_tmp[ ivI ].Start( seqI ) );
			gnSeqI right_end = left_end + iv_list_tmp[ ivI ].Length( seqI ) - 1;
			island_out << "LCB island:\t" << seqI << '\t' << left_end << '\t' << right_end << endl;
		}
	}
}


void allocateDetailList( MatchList& mlist, vector< pair< uint64, uint64 > >& detail_list ){
	IntervalList iv_list;
	iv_list.seq_table = mlist.seq_table;
	iv_list.seq_filename = mlist.seq_filename;
	allocateDetailList( iv_list, detail_list );
}

void getLCBDetailList( MatchList& mlist, vector< pair< uint64, uint64 > >& detail_list ){
	Interval iv( mlist.seq_table.size() );
	iv.matches.insert( iv.matches.end(), mlist.begin(), mlist.end() );
	getLCBDetailList( iv, detail_list );

}


void allocateDetailList( IntervalList& iv_list, vector< pair< uint64, uint64 > >& detail_list ){
	uint seq_count = iv_list.seq_table.size();
	uint64 max_types = 1 << seq_count; // (uint64) pow( (double)2, (double)seq_count );
	detail_list.clear();
	uint64 seqI;
	for( seqI = 0; seqI < max_types; seqI++ )
		detail_list.push_back( pair< uint64, uint64 >( seqI, 0 ) );

	uint64 match_bits = 1;
	for( seqI = seq_count; seqI > 0; seqI-- ){
		detail_list[ match_bits ].second = iv_list.seq_table[ seqI - 1 ]->length();
		match_bits <<= 1;
	}
}

void getLCBDetailList( Interval& iv, vector< pair< uint64, uint64 > >& detail_list ) {
	if( iv.matches.size() == 0 )
		return;
	uint seq_count = iv.matches[ 0 ]->SeqCount();
	uint seqI;
	for( uint matchI = 0; matchI < iv.matches.size(); matchI++ ){
		//get the sequence matching entry into match bits
		const GappedAlignment* cr = dynamic_cast< const GappedAlignment* >( iv.matches[ matchI ] );
		const Match* match = dynamic_cast< const Match* >( iv.matches[ matchI ] );
		if( match != NULL ){
			uint64 match_bits = 0;
			for( seqI = 0; seqI < seq_count; seqI++ ){
				match_bits <<= 1;
				if((*match)[ seqI ] != MEM_NO_MATCH){
					match_bits |= 0x1;
					uint64 cur_matchnum = 1;
					cur_matchnum <<= (seq_count - seqI - 1);
					detail_list[ cur_matchnum ].second -= match->Length();
				}
			}
			detail_list[ match_bits ].second += match->Length();
		}else if( cr != NULL ){
			for( gnSeqI charI = 0; charI < cr->align_length; charI++ ){
				for( seqI = 0; seqI < seq_count; seqI++ ){
					uint64 cur_matchnum = 0;
					uint seqJ = 0;
					uint64 match_bits = 1;
					match_bits <<= seqI;
					boolean matched = false;
					for( seqJ = 0; seqJ < seq_count; seqJ++ ){
						if( seqJ > seqI )
							match_bits <<= 1;
						if( toupper( cr->align_matrix[ seqI ][ charI ] ) == 
							toupper( cr->align_matrix[ seqJ ][ charI ] ) )
							if( seqI < seqJ && cr->align_matrix[ seqI ][ charI ] != '-' ){
								cur_matchnum = 1;
								cur_matchnum << seq_count - seqJ - 1;
								detail_list[ cur_matchnum ].second -= 1;
								match_bits |= 0x1;
								matched = true;
							}else if( seqI > seqJ )
								break;
						
					}
					if( seqJ == seq_count && matched ){
						cur_matchnum = 1;
						cur_matchnum <<= seq_count - seqI - 1;
						detail_list[ cur_matchnum ].second -= 1;
						detail_list[ match_bits ].second += 1;
					}
						
				}
			}
		}
	}
}


void PrintDetailList( uint seq_count, const vector< pair< uint64, uint64 > >& detail_list, ostream& os, boolean skip_zeros ){
	uint64 detailMax = (uint64) pow((double)2, (double)seq_count);
	for(uint64 detailI = 0; detailI < detailMax; detailI++){
		string seq_string;
		if( skip_zeros && detail_list[detailI].second == 0 )
			continue;
		uint64 curDetail = detail_list[ detailI ].first;
		for(uint32 seqI = 0; seqI < seq_count; seqI++){
			if(curDetail & 0x1)
				seq_string = "1 " + seq_string;
			else
				seq_string = "0 " + seq_string;
			curDetail >>= 1;
		}
		os << seq_string << '\t' << detail_list[detailI].second << '\n';
	}
}

void IdentityMatrix( IntervalList& iv_list, NumericMatrix<double>& identity ){
	if( iv_list.size() == 0 )
		return;

	uint seq_count = iv_list.seq_table.size();
	identity = NumericMatrix<double>( seq_count, seq_count );
	identity.init( 0 );
	NumericMatrix<double> possible( seq_count, seq_count );
	possible.init( 0 );
	
	for( uint ivI = 0; ivI < iv_list.size(); ivI++ ){
		NumericMatrix<double> cur_id;
		IdentityMatrix( iv_list[ ivI ], cur_id );
		for( uint seqI = 0; seqI < seq_count; seqI++ ){
			for( uint seqJ = 0; seqJ < seq_count; seqJ++ ){
				gnSeqI shorter_len = iv_list[ ivI ].Length( seqI ) < iv_list[ ivI ].Length( seqJ ) ? iv_list[ ivI ].Length( seqI ) : iv_list[ ivI ].Length( seqJ );
				cur_id( seqI, seqJ ) *= (double)shorter_len;
				possible( seqI, seqJ ) += shorter_len;
			}
		}
		identity += cur_id;
	}
	identity /= possible;
}

void IdentityMatrix( Interval& iv, NumericMatrix<double>& identity ){
	if( iv.matches.size() == 0 )
		return;
	uint seq_count = iv.matches[ 0 ]->SeqCount();
	identity = NumericMatrix<double>( seq_count, seq_count );
	identity.init( 0 );
	uint seqI;
	uint seqJ;

	for( uint matchI = 0; matchI < iv.matches.size(); matchI++ ){
		//get the sequence matching entry into match bits
		const GappedAlignment* cr = dynamic_cast< const GappedAlignment* >( iv.matches[ matchI ] );
		const Match* match = dynamic_cast< const Match* >( iv.matches[ matchI ] );
		if( match != NULL ){
			for( seqI = 0; seqI < seq_count; seqI++ ){
				for( seqJ = seqI + 1; seqJ < seq_count; seqJ++ ){
					identity( seqI, seqJ ) += match->Length();
				}
			}
		}else if( cr != NULL ){
			for( gnSeqI charI = 0; charI < cr->align_length; charI++ ){
				for( seqI = 0; seqI < seq_count; seqI++ ){
					for( seqJ = seqI + 1; seqJ < seq_count; seqJ++ ){
						if( ( toupper( cr->align_matrix[ seqI ][ charI ] ) == 
							toupper( cr->align_matrix[ seqJ ][ charI ] ) ) &&
							cr->align_matrix[ seqI ][ charI ] != '-' ){
							
								identity( seqI, seqJ ) += 1;
						}
					}
				}
			}
		}
	}

	for( seqI = 0; seqI < seq_count; seqI++ ){
		for( seqJ = seq_count; seqJ > 0; seqJ-- ){
			if( seqI == seqJ - 1 )
				// set the diagonal to identical
				identity( seqI, seqJ - 1 ) = 1;
			else if( seqI < seqJ - 1 ){
				// determine the length of the shorter sequence
				gnSeqI shorter_len = iv.Length( seqI ) < iv.Length( seqJ - 1 ) ? iv.Length( seqI ) : iv.Length( seqJ - 1 );
				// divide through
				identity( seqI, seqJ - 1 ) /= (double)shorter_len;
			}else	// copy the other one
				identity( seqI, seqJ - 1 ) = identity( seqJ - 1, seqI );
		}
	}
	
}

void DistanceMatrix( uint seq_count, const vector< pair< uint64, uint64 > >& detail_list, NumericMatrix<double>& distance ){
	distance = NumericMatrix<double>( seq_count, seq_count );
	distance.init( 0 );
	uint seqI;
	uint seqJ;
	for( seqI = 0; seqI < seq_count; seqI++ ){
		uint64 seqI_mask = 1;
		seqI_mask <<= seq_count - seqI - 1;
		for( seqJ = 0; seqJ < seq_count; seqJ++ ){
			uint64 seqJ_mask = 1;
			seqJ_mask <<= seq_count - seqJ - 1;
			for( uint pairI = 0; pairI < detail_list.size(); pairI++ ){
				if( (detail_list[ pairI ].first & seqI_mask) != 0 &&
					(detail_list[ pairI ].first & seqJ_mask) != 0 ){
					distance( seqI, seqJ ) += detail_list[ pairI ].second;
				}
			}
		}
	}
	
	for( seqI = 0; seqI < seq_count; seqI++ ){
		for( seqJ = 0; seqJ < seq_count; seqJ++ ){
			if( seqI == seqJ )
				continue;
			double avg_length = ( distance( seqI, seqI ) + distance( seqJ, seqJ ) ) / 2;
			distance( seqI, seqJ ) = 1.0 - ( distance( seqI, seqJ ) / avg_length );
			if( !(distance( seqI, seqJ ) == distance( seqI, seqJ )) ){
				distance( seqI, seqJ ) = 1.0;
			}
		}
	}

	// set the diagonal identical to itself
	for( seqI = 0; seqI < seq_count; seqI++ )
		distance( seqI, seqI ) = 0;
}


/*
void Aligner::GetLCBIdentityMatrix( MatchList& lcb, NumericMatrix< double >& identity, NumericMatrix< double >& range ){
	NumericMatrix< double > density( seq_count, seq_count );
	NumericMatrix< double > m_range( seq_count, seq_count );
	NumericMatrix< double > last_coord( seq_count, seq_count );
	uint seqI, seqJ;
	density.init( 0 );
	m_range.init( 0 );
	last_coord.init( 0 );

	vector< Match* >::iterator match_iter = lcb.begin();
	for( ; match_iter != lcb.end(); match_iter++ ){
		for( seqI = 0; seqI < seq_count; seqI++ ){
			if( (*match_iter)->Start( seqI ) == MEM_NO_MATCH )
				continue;
			for( seqJ = 0; seqJ < seq_count; seqJ++ ){
				if( (*match_iter)->Start( seqJ ) == MEM_NO_MATCH || seqI == seqJ )
					continue;
				last_coord( seqI, seqJ ) = (*match_iter)->End( seqI );
				density( seqI, seqJ ) += (*match_iter)->Length();
				if( m_range( seqI, seqJ ) == 0 ){
					m_range( seqI, seqJ ) = (*match_iter)->Start( seqI );
					if( m_range( seqI, seqJ ) < 0 )
						m_range( seqI, seqJ ) -= (*match_iter)->Length() - 1;
				}
			}
		}
	}
	for( seqI = 0; seqI < seq_count; seqI++ ){
		for( seqJ = 0; seqJ < seq_count; seqJ++ ){
			if( seqI == seqJ )
				density( seqI, seqJ ) = 1;
			else if( last_coord( seqI, seqJ ) != 0 ){
				m_range( seqI, seqJ ) = abs( last_coord( seqI, seqJ ) - m_range( seqI, seqJ ) ) + 1;
				density( seqI, seqJ ) = density( seqI, seqJ ) / (double)m_range( seqI, seqJ );
			}
		}
	}
	range = m_range;
	identity = density;
}
*/
